from .compat import hex2bytes, urlsplit


def sz(x):
    s = hex(x if isinstance(x, int) else len(x))[2:].rjust(4, '0')
    s = hex2bytes(s)
    return s[::-1]


def pack_uwsgi_vars(var):
    pk = b''
    for k, v in var.items() if hasattr(var, 'items') else var:
        pk += sz(k) + k.encode('utf8') + sz(v) + v.encode('utf8')
    return b'\x00' + sz(pk) + b'\x00' + pk


def parse_addr(addr, default_port=3030):
    host = None
    port = None
    if isinstance(addr, str):
        if addr.isdigit():
            port = addr
        else:
            parts = urlsplit('//' + addr)
            host = parts.hostname
            port = parts.port
    elif isinstance(addr, (list, tuple, set)):
        host, port = addr
    return (host or '127.0.0.1',
            int(port) if port else default_port)


def get_host_from_url(url):
    # TODO: error for https
    url = url.split('://')[-1]

    if url and url[0] != '/':
        # TODO: validate hostname
        host, _, url = url.partition('/')
        return (host, '/' + url)

    return '', url



import socket
import sys
from .compat import urlsplit
from .utils import pack_uwsgi_vars, parse_addr, get_host_from_url


def ask_uwsgi(uwsgi_addr, var, body='', timeout=0, udp=False):
    sock_type = socket.SOCK_DGRAM if udp else socket.SOCK_STREAM
    if isinstance(uwsgi_addr, str) and '/' in uwsgi_addr:
        addr = uwsgi_addr
        s = socket.socket(family=socket.AF_UNIX, type=sock_type)
    else:
        addr = parse_addr(addr=uwsgi_addr)
        s = socket.socket(*socket.getaddrinfo(
            addr[0], addr[1], 0, sock_type)[0][:2])

    if timeout:
        s.settimeout(timeout)

    if body is None:
        body = ''

    s.connect(addr)
    s.send(pack_uwsgi_vars(var) + body.encode('utf8'))
    response = []
    while 1:
        data = s.recv(4096)
        if not data:
            break
        response.append(data)
    s.close()
    return b''.join(response).decode('utf8')


def curl(uwsgi_addr, url, method='GET', body='', timeout=0, headers=(),
         udp=False):
    host, uri = get_host_from_url(url)
    parts_uri = urlsplit(uri)

    if '/' not in uwsgi_addr:
        addr = parse_addr(addr=uwsgi_addr)
        if not host:
            host = addr[0]
        port = addr[1]
    else:
        port = None

    var = {
        'SERVER_PROTOCOL': 'HTTP/1.1',
        'PATH_INFO': parts_uri.path,
        'REQUEST_METHOD': method.upper(),
        'REQUEST_URI': uri,
        'QUERY_STRING': parts_uri.query,
        'HTTP_HOST': host,
    }
    for header in headers or ():
        key, _, value = header.partition(':')
        var['HTTP_' + key.strip().upper()] = value.strip()
    var['SERVER_NAME'] = var['HTTP_HOST']
    if port:
        var['SERVER_PORT'] = str(port)
    result = ask_uwsgi(uwsgi_addr=uwsgi_addr, var=var, body=body,
                       timeout=timeout, udp=udp)
    return result


def cli(*args):
    import argparse

    parser = argparse.ArgumentParser()

    parser.add_argument('uwsgi_addr', nargs=1,
                        help="Remote address of uWSGI server")

    parser.add_argument('url', nargs='?', default='/',
                        help="Request URI optionally containing hostname")

    parser.add_argument('-X', '--method', default='GET',
                        help="Request method. Default: GET")

    parser.add_argument('-H', '--header', action='append', dest='headers',
                        help="Request header. It can be used multiple times")

    parser.add_argument('-d', '--data', help="Request body")

    parser.add_argument('-t', '--timeout', default=0.0, type=float,
                        help="Socket timeout")

    parser.add_argument('--udp', action='store_true',
                        help="Use UDP instead of TCP")

    args = parser.parse_args(args or sys.argv[1:])

    response = curl(uwsgi_addr=args.uwsgi_addr[0], method=args.method,
                    url=args.url, body=args.data, timeout=args.timeout,
                    headers=args.headers, udp=args.udp)
    print(response)

    status = int(response.split(' ', 2)[1])
    return not (200 <= status < 300)


if __name__ == '__main__':
    sys.exit(cli(*sys.argv[1:]))



import sys
from wsgiref.simple_server import WSGIRequestHandler

from .compat import BaseHTTPRequestHandler, TCPServer, get_content_type
from .curl import ask_uwsgi
from .utils import parse_addr


class RequestHandler(WSGIRequestHandler):
    handle = BaseHTTPRequestHandler.handle

    def write(self, s):
        self.wfile.write(s.encode('utf8'))
        self.close_connection = True

    def do(self):
        self.log_message(self.requestline.rpartition(' ')[0]
                         .replace('%', '%%'))

        if self.server.redirect_static:
            if any(map(self.path.startswith, ['/static/', '/media/'])):
                return self.write('HTTP/1.1 302 Found\nLocation: http://%s%s\n'
                                  % (self.server.uwsgi_host, self.path))

        setattr(self.server, 'base_environ', {})
        env = WSGIRequestHandler.get_environ(self)
        qs = env['QUERY_STRING']
        env = dict(dict({
            'REQUEST_URI': env['PATH_INFO'] + (('?' + qs) if qs else ''),
            'CONTENT_LENGTH': '',
            'DOCUMENT_ROOT': '',
            'REMOTE_PORT': str(self.client_address[1]),
        }, **env), **{
            'SERVER_ADDR': self.server.uwsgi_addr,
            'SERVER_NAME': self.server.uwsgi_host,
            'SERVER_PORT': str(self.server.uwsgi_port),
            'CONTENT_TYPE': get_content_type(self.headers) or '',
            'HTTP_HOST': self.server.uwsgi_host,
        })
        env.pop('REMOTE_HOST', 0)
        env.pop('HTTP_REFERER', 0)

        cl = env['CONTENT_LENGTH']
        body = repr(self.rfile.read(int(cl))) if cl else ''

        resp = ask_uwsgi((self.server.uwsgi_addr, self.server.uwsgi_port),
                         var=env, body=body)
        self.write(resp)
        h, _, b = resp.partition('\r\n\r\n')
        print('%s\n%s' % (h, len(b)))

    do_HEAD = do_GET = do_POST = do


def serve_forever(uwsgi_addr, uwsgi_host=None, local_addr='',
                  redirect_static=True):

    uwsgi_addr, uwsgi_port = parse_addr(uwsgi_addr, 3030)
    local_addr = parse_addr(local_addr, uwsgi_port)

    print('Proxying remote uWSGI server %s:%s "%s" to local HTTP server '
          '%s:%s...' % (uwsgi_addr, uwsgi_port, (uwsgi_host or ''),
                        local_addr[0], local_addr[1]))

    TCPServer.allow_reuse_address = True
    s = TCPServer(
        server_address=parse_addr(local_addr),
        RequestHandlerClass=RequestHandler,
    )
    s.uwsgi_addr = uwsgi_addr
    s.uwsgi_port = uwsgi_port
    s.uwsgi_host = uwsgi_host or uwsgi_addr
    s.redirect_static = redirect_static
    try:
        s.serve_forever()
    except (KeyboardInterrupt, SystemExit):
        print(' Bye.')
    finally:
        s.shutdown()
        s.server_close()


def cli(*args):
    import argparse

    parser = argparse.ArgumentParser()

    parser.add_argument('uwsgi_addr', nargs=1,
                        help='Remote address of uWSGI server')
    parser.add_argument('-n', '--http-host',
                        help='HTTP_HOST header for uWSGI server')
    parser.add_argument('-l', '--local-addr', default='127.0.0.1',
                        help='Local address of HTTP server')
    parser.add_argument('-s', '--redirect-static', action='store_true',
                        default=False, help='Redirect static requests to host')

    args = parser.parse_args(args or sys.argv[1:])
    serve_forever(
        uwsgi_addr=args.uwsgi_addr[0],
        uwsgi_host=args.http_host,
        local_addr=args.local_addr,
        redirect_static=args.redirect_static,
    )


if __name__ == '__main__':
    cli()



import sys

__all__ = [
    'BaseHTTPRequestHandler', 'TCPServer', 'get_content_type', 'urlsplit',
    'hex2bytes',
]

PY3 = sys.version_info[0] == 3


if PY3:
    from http.server import BaseHTTPRequestHandler
    from socketserver import TCPServer
    from urllib.parse import urlsplit
else:
    from BaseHTTPServer import BaseHTTPRequestHandler
    from SocketServer import TCPServer
    from urlparse import urlsplit


def get_content_type(headers):
    if PY3:
        return headers.get('content-type')
    else:
        return headers.typeheader


def hex2bytes(s):
    if PY3:
        return bytes.fromhex(s)
    else:
        return s.decode('hex')



from setuptools import setup

version = '1.1.1'

setup(
    name='uwsgi-tools',
    version=version,
    author='Andrei Fokau',
    author_email='andrei@5monkeys.se',
    url='https://github.com/andreif/uwsgi-tools',
    download_url='https://github.com/andreif/uwsgi-tools/tarball/%s' % version,
    license='MIT',
    description='uwsgi tools: curl and reverse proxy',
    long_description='''
        Reverse proxy and curl for uWSGI server using `uwsgi` protocol
    ''',
    classifiers=[
        'Environment :: Web Environment',
        'Natural Language :: English',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    packages=['uwsgi_tools'],
    entry_points={
        'console_scripts': [
            'uwsgi_proxy = uwsgi_tools.proxy:cli',
            'uwsgi_curl = uwsgi_tools.curl:cli',
        ]
    },
    install_requires=[],
    tests_require=[],
    test_suite='runtests.main'
)
